package dataspork.stat;
import dataspork.util.DoubleWindowModel;
import dataspork.util.DataListener;
import java.util.EventListener;
import javax.swing.event.EventListenerList;

/** Class for representing time series data.  Calculates mean and variance,
  * and allows for imin and imax cutoffs.
  */
public class ScalarSeries implements DataListener{

  /** The mean. */
  protected double mean=0d;

  /** The variance. */
  protected double var=0d;

  /** The standard deviation. */
  protected double sigma=0d;

  /** The autocorrelation time. */
  protected double autocor=1d;

  /** The cutoff window for calculating statistics. */
  protected DoubleWindowModel window;

  /** The original time series data */
  public double fullData[]=null;

  /** The current blocking factor */
  public int block;

  /** The blocked time series data */
  public double data[];

  /** Boolean flag for non-zero time series */
  public boolean hasData;

  /** Data listeners for this object */
  protected EventListenerList listenerList;
  
  /** Create a ScalarSeries object that refers to the given array.
    */
  public ScalarSeries(double[] data) {
    if (data.length > 1) {
      this.hasData = true;
      this.data = data;
      window = new DoubleWindowModel(0,data.length-1,1);
      listenerList = new EventListenerList();
      calcMean();
      window.addListener(this);
    }
    else {
      this.hasData = false;
      this.data = null;
      window = new DoubleWindowModel(0,0,1);
      listenerList = new EventListenerList();
      window.addListener(this);
    }
  }

  /** Set the autocorrelation time.  Used for calculating the error
    * of the average.
    */ 
  public void setAutocor(double time) {
    autocor=(time>1)?time:1;
  }
  
  /** Return the data.
    */
  public double[] getData() {
    return data;
  }

  /** Block the data, return the new array.
    */
  public double[] blockData() {
    int nblock = window.getBlock();

    if (fullData==null) { fullData = data; }

    if (nblock==0) {
      data = fullData;
    }
    else {
      int npoints = (int)Math.pow(2,nblock);
      int templength = fullData.length/npoints;
      double[] temp = new double[templength];
      int tmpcnt = 0;
      for (int i=0; i<temp.length; i++) {
        double tempsum = 0.;
        for (int j=0; j<npoints; j++) { tempsum += fullData[tmpcnt++]; }
        temp[i] = tempsum/npoints;
      }
      data = temp;
    }
    return data;
  }

  /** Return the mean.
    */
  public double getMean() {
    return mean;
  }

  /** Return the error of the mean.
    */
  public double getErrorOfMean() {
    if (window.getHigh() > window.getLow()) {
      return sigma*Math.sqrt(autocor/(window.getHigh()-window.getLow()));
    }
    else {
      return 0.;
    }
  }

  /** Return the variance.
    */
  public double getVariance() {
    return var;
  }

  /** Return the standard deviation.
    */
  public double getSigma() {
    return sigma;
  }

  /** Return the autocorrelation time being used for estimating
    * the error of the mean.  This value is not calculated by
    * this class, but can be set using setCorrTime.
    */
  public double getAutoCorr() {
    return autocor;
  }

  /** Return the model for the datarange.  */
  public DoubleWindowModel getWindow() {
    return window;
  }

  protected void calcMean() {
    if (block != window.getBlock()) { 
      blockData();
      block = window.getBlock();
    }

    mean = 0d;
    var = 0d;
    int imin = (int)window.getLow();
    int imax = (int)window.getHigh();
    if (imax>data.length-1) { imax=data.length-1; }
    int n=imax-imin+1;

    if (n <= 0) {
      System.out.println("Problem in ScalarSeries!"+n);
    }
    else {
      for (int i=imin; i<=imax; i++) {
        mean+=data[i];
        var+=data[i]*data[i];
      }
      mean/=n;
      var = var/n - mean*mean;
    }
    sigma = Math.sqrt(var);
  }

  public void addListener(DataListener e) {
    listenerList.add(DataListener.class,e);
  }

  protected void fireListeners(Object o) {
    Object[] listeners = listenerList.getListenerList();
    for (int i = listeners.length-2; i>=0; i-=2) {
      if (o!=listeners[i+1]) {
        ((DataListener)listeners[i+1]).dataChanged(this);
      }
    }
  }

  public void dataChanged(Object o) {
    calcMean();
    fireListeners(o);
  }

}
