package dataspork.util;

import javax.swing.event.EventListenerList;
import java.util.EventListener;


/** Model for storing a range low, high, with constraints.
  * Require that high - low is greater than or equal the window size width.
  * Also constrains with an min and max.
  */
public class DoubleWindowModel {
  
  protected double low, high;
  protected double min, max;
  protected double width;
  protected double fullMin, fullMax;
  protected int block;
  protected EventListenerList listenerList;

  public DoubleWindowModel(double min, double max, double width) {
    low=min;
    high=max;
    this.min=min;
    this.max=max;
    this.width=width;
    this.block=0;
    this.fullMin = min;
    this.fullMax = max;
    listenerList = new EventListenerList();
  }

  /** Return the lower limit of the current range. */
  public double getLow() {
    return low;
  }

  /** Return the upper limit of the current range. */
  public double getHigh() {
    return high;
  }

  /** Return the upper limit of the current range. */
  public int getBlock() {
    return block;
  }

  /** Attempt to change the lower value. */
  public void setLow(double lo) {
    if      (lo<min)        {lo=min;}
    else if (lo>high-width) {lo=high-width;}
    if (lo!=low) {
      low=lo;
      fireListeners();
    }
  }

  /** Attempt to change the upper value. */
  public void setHigh(double hi) {
    if      (hi>max)       {hi=max;}
    else if (hi<low+width) {hi=low+width;}
    if (hi!=high) {
      high=hi;
      fireListeners();
    }
  }

  /** Attempt to change the blocking factor. */
  public void setBlock(int blockingFactor) {
    int maxblock=(int)Math.floor(Math.log(fullMax-fullMin)
                 /Math.log(2d))-2;
    if (blockingFactor>maxblock) { blockingFactor=maxblock; }
    if (blockingFactor<0) { blockingFactor=0; }

    if (blockingFactor!=block) {
      // ADJUST HI AND MAX TO NEW BLOCKING
      int npoints = (int)Math.pow(2,blockingFactor);
      int oldnpoints = (int)Math.pow(2,block);
      if (npoints > 0) {
        max = (int)fullMax/npoints - 1;
        high = (int)high*oldnpoints/npoints;
        if (high>max) high=max;
        low = (int)low*oldnpoints/npoints;
      }
      block=blockingFactor;
      fireListeners();
    }
  }

  public void addListener(DataListener e) {
    listenerList.add(DataListener.class,e);
  }

  protected void fireListeners() {
    Object[] listeners = listenerList.getListenerList();
    for (int i = listeners.length-2; i>=0; i-=2) {
      ((DataListener)listeners[i+1]).dataChanged(this);
    }
  }
}
