/****************************************************************************
 * NCSA HDF                                                                 *
 * National Comptational Science Alliance                                   *
 * University of Illinois at Urbana-Champaign                               *
 * 605 E. Springfield, Champaign IL 61820                                   *
 *                                                                          *
 * For conditions of distribution and use, see the accompanying             *
 * hdf/COPYING file.                                                        *
 *                                                                          *
 ****************************************************************************/

package ncsa.hdf.message;

import java.util.Vector;

import ncsa.hdf.util.Queue;
import ncsa.hdf.hdflib.*;

/* This class will analyse the HDF file and put the HDF object into 
 * the Queue. Hdf object is the node created by HDFObjectNode class
 * Date: 3-12-97 Modified "getSDSObject()".
 *    Keep the name of an HDF object node to be more readable,
 *    to be as the form look like "sdsname: long_name".
 * @version 2.00  9/3/97
 * @auther  Xinjian Lu, modified by Peter Cao on 4-14-98
 */
public class HDFAnalyse
{
    /** HDF Library */
    private HDFLibrary hdf;

    /** file id */
    private int fid = -1;

    /** hdf file name */
    private String fileName = "";

    /** the display mode */
    private boolean debug = false;


    /** analyse the HDF file then put it into Queue
     * @param hdffile   the HDF file name
     * @param hdfQueue  the queue to hold the HDF objects
     */
    public void getHdfObject(String hdffile, Queue hdfObjQueue)
    {
        this.hdf = new HDFLibrary();
	try
        {
            if ((fid = hdf.Hopen(hdffile)) <= 0)
                return;
            this.fileName = hdffile;

            // another HDFObjectNode variable
            HDFObjectNode   parentNode, childNode, node;

            // for HDF annotation;
            if (ANhasFileAnn(fid))
            {
                // HDF Object Annotation root node
                node = new HDFObjectNode();
                node.setObjType(node.Annotation);
                node.setName("File Annotation");
                hdfObjQueue.put(node);
            }

            // for HDF Objects, use this order of calls to
            // track if a data object is lone or not
            getVgroupObject(fid, hdfObjQueue);
            getVdataObject(fid, hdfObjQueue);
            getSDSObject(hdffile, hdfObjQueue);
            getGRObject(fid, hdfObjQueue);
            getR24Object(hdffile,hdfObjQueue );
            getR8Object(hdffile,hdfObjQueue );

            hdf.Hclose(fid);
        } catch (HDFException e) {}
    }

    /** put the 8-bit raster image object into HDF queue
     * @param filenae  HDF file name
     * @param objQueue the queue to store the HDF object node
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     * @see HDFObjectNode
     */
    public void getR8Object(String filename, Queue objQueue)
        throws HDFException
    {
        //System.out.println("HDFAnalyse.getR8Object()");
        
        String info = "";
        int tag, ref, imgNumber, loneImageNumber=0;
        HDFObjectNode parentNode, childNode, node;
        int argv[] = new int[2];
        boolean argB[] = new boolean[1];

        if(  !hdf.DFR8restart() ||
            (imgNumber=hdf.DFR8nimages(filename)) <=0 )
            return;

        // HDF Object RIS8 root node
        node = new HDFObjectNode();
        node.setObjType(node.RIS8);
        node.setObjLevel(0);
        node.setName("8-bit Raster Image");
        node.setHasChild(true);
        Queue childQueue = new Queue();
        node.setChildObj(childQueue);
        parentNode = node;

        for (int i=0; i<imgNumber; i++)
        {
            ref = hdf.DFR8lastref();
            tag = HDFConstants.DFTAG_RIG;

            if (!hdf.DFR8readref(filename, ref) ||
                ref == 0 || find(tag, ref, objQueue) )
                continue;

            hdf.DFR8getdims(filename,  argv, argB);
            info  = "\nTag number           : "+tag;
            info += "\nReference number     : "+ref;
            info += "\nImage size           : "+argv[0]+" X "+argv[1];
            info += "\nHas palette          : "+argB[0];

            // HDF Object RIS8 root node
            node = new HDFObjectNode();
            node.setObjType(node.RIS8);
            node.setTagRef(tag,ref);
            node.setObjLevel(1);
            node.setName("Raster Image #" + i);
            node.setHasParent(true);
            node.setParentObj(parentNode);
            node.information = info;
            childQueue.put(node);

            loneImageNumber++;
        }

        if (loneImageNumber > 0)
        {
            parentNode.information = "\nNumber of raster images : "+loneImageNumber;
            objQueue.put(parentNode);
        }
    }

    /** put the 24-bit raster image object into HDF queue
     * @param filenae  HDF file name
     * @param objQueue the queue to store the HDF object node
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     * @see HDFObjectNode
     */
    public void getR24Object(String filename, Queue objQueue)
        throws HDFException
    {
        String info = "";
        int args[] = new int[3];
        int tag, ref, imgNumber, loneImageNumber=0;
        HDFObjectNode parentNode, childNode, node;

        if (!hdf.DF24restart() ||
            (imgNumber=hdf.DF24nimages(filename)) <= 0 )
            return;

        // HDF Object RIS24 root node
        node = new HDFObjectNode();
        node.setObjType(node.RIS24);
        node.setObjLevel(0);
        node.setName("24-bit Raster Image");
        node.setHasChild(true);
        Queue childQueue = new Queue();
        node.setChildObj(childQueue);
        parentNode = node;

        for (int i=0; i<imgNumber; i++)
        {
            ref = hdf.DF24lastref();
            tag = HDFConstants.DFTAG_RIG;
            hdf.DF24getdims(filename, args);

            if (ref == 0 || find(tag, ref, objQueue))
                continue;

            info  = "\nTag number           : "+tag;
            info += "\nReference number     : "+ref;
            info += "\nImage size           : "+args[0]+" X "+args[1];

            // HDF Object RIS24 root node
            node = new HDFObjectNode();
            node.setObjType(node.RIS24);
            node.setTagRef(tag,ref);
            node.setObjLevel(1);
            node.setName("Raster Image #"+i);
            node.setHasParent(true);
            node.setParentObj(parentNode);
            node.information = info;
            childQueue.put(node);

            loneImageNumber++;
        }

        if (loneImageNumber > 0)
        {
            parentNode.information = "\nNumber of raster images : "+loneImageNumber;
            objQueue.put(parentNode);
        }
    }

    /** put the Vgroup object into HDF queue
     * @param fid  HDF file identifier by Hopen
     * @param objQueue the queue to store the HDF object node
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     * @see HDFObjectNode
     */
    public void getVgroupObject(int fid, Queue objQueue)
        throws HDFException
    {
        HDFObjectNode parentNode, childNode, node;
        int[] n = new int[1];

        hdf.Vstart(fid);

        // get the total number of lone vgroups
        int vgNumber = hdf.Vlone(fid,n,0);
        if (vgNumber <= 0)
        {
            hdf.Vend(fid);
            return;
        }

        int refArray[] = new int[vgNumber];
        int refNum = hdf.Vlone(fid, refArray, vgNumber);

        if (vgNumber <= 0)
        {
            hdf.Vend(fid);
            return;
        }

        int tags[] = new int[100];
        int refs[] = new int[100];
        String info = "";
        for (int i=0; i<refNum; i++)
        {
            int vgroup_ref = refArray[i];
            int vgtag = HDFConstants.DFTAG_VG;
            int vgroup_id = hdf.Vattach(fid, vgroup_ref, "r");
            if (vgroup_id == HDFConstants.FAIL)
            {
                hdf.Vdetach(vgroup_id);
                continue;
            }

            String s[] = {""};
            hdf.Vgetname(vgroup_id,s);
            String vgName = s[0].trim();
            hdf.Vgetclass(vgroup_id,s);
            String vgclass = s[0].trim();
            int npairs = hdf.Vgettagrefs(vgroup_id, tags, refs, 100);

            if ( vgclass.equalsIgnoreCase(HDFConstants.GR_NAME) ||
                 vgclass.equalsIgnoreCase(HDFConstants.HDF_CDF) ||
                 (npairs <= 0) )
                continue;

            info  = "\nVgroup Name            : "+vgName;
            info += "\nTag number             : "+vgtag;
            info += "\nReference number       : "+vgroup_ref;
            info += "\nNumber of data objects : "+npairs;

            node = new HDFObjectNode();
            node.setObjType(node.Vgroup);
            node.setTagRef(vgtag, vgroup_ref);
            node.setName("Vgroup ("+vgName + ")");
            node.setHasChild(true);
            Queue vgChildQueue = new Queue(npairs);
            node.setChildObj(vgChildQueue);
            node.information = info;
            objQueue.put(node);

            insertVg(node, vgChildQueue, fid, tags, refs, npairs, 1);

            hdf.Vdetach(vgroup_id);
        }

        hdf.Vend(fid);
    }

    /** put the Vdata object into HDF queue
     * @param fid  HDF file identifier by Hopen
     * @param objQueue the queue to store the HDF object node
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     * @see HDFObjectNode
     */
    public void getVdataObject(int fid, Queue objQueue)
        throws HDFException
    {
        HDFObjectNode   parentNode, childNode, node;
        String info = "";
        int numberOfVdata = 0;

        hdf.Vstart(fid);

        // for Vdata
        int [] n = new int[1];
        int vdNumber = hdf.VSlone(fid,n,0);

        if (vdNumber <= 0)
        {
            hdf.Vend(fid);
            return;
        }

        // Vdata object root node
        node = new HDFObjectNode();
        node.setObjType(node.Vdata);
        node.setName("Vdata (lone)");

        node.setHasChild(true);
        parentNode = node;
        Queue vdQueue = new Queue(vdNumber);
        parentNode.setChildObj(vdQueue);

        int refArray[] = new int[vdNumber];
        int refNum = hdf.VSlone(fid, refArray, vdNumber);

        if (refNum <= 0)
        {
            hdf.Vend(fid);
            return;
        }

        for (int i=0; i<refNum; i++)
        {
            String[] vdName = {""};
            String[] vdClass = {""};
            int vdata_ref = refArray[i];
            int vdtag = HDFConstants.DFTAG_VS;
            int vdata_id = hdf.VSattach(fid, vdata_ref, "r");

            hdf.VSgetclass(vdata_id, vdClass);
            if (!debug && vdClass!=null &&
                vdClass[0].length()>=HDFConstants.HDF_CHK_TBL.length() &&
                vdClass[0].substring(0,HDFConstants.HDF_CHK_TBL.length()).equalsIgnoreCase(
                HDFConstants.HDF_CHK_TBL))
                continue;

            hdf.VSgetname(vdata_id,vdName);

            // add node information
            info  = "\nVdata Name        : "+vdName[0];
            info += "\nTag number        : "+vdtag;
            info += "\nReference number  : "+vdata_ref;
            info += "\nNumber of records : "+hdf.VSelts(vdata_id);
            info += "\nNumber of fields  : "+hdf.VFnfields(vdata_id);

            node = new HDFObjectNode();
            node.setObjType(node.Vdata);
            node.setTagRef(vdtag, vdata_ref);
            node.setName(vdName[0]);
            node.setHasParent(true);
            node.setParentObj(parentNode);
            node.setObjLevel(1);
            node.information = info;
            vdQueue.put(node);

            hdf.VSdetach(vdata_id);

            numberOfVdata++;
        }

        if ( numberOfVdata > 0)
        {
            info = "\nNumber of lone vdatas : "+numberOfVdata;
            parentNode.information = info;
            objQueue.put(parentNode);
        }

	hdf.Vend(fid);
    }

    /** put the SDS object into HDF queue
     * @param filename  HDF file name
     * @param objQueue the queue to store the HDF object node
     *
     *  @exception ncsa.hdf.hdflib.HDFException
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     * @see HDFObjectNode
     */
    public void getSDSObject(String filename, Queue objQueue)
        throws HDFException
    {
        int tag, ref;
        HDFObjectNode   parentNode, childNode, node;
        int[] fileInfo = new int[2];
        int sdid = hdf.SDstart(filename, HDFConstants.DFACC_RDONLY);
        String info = "";
        int numberOfData = 0;

        if( sdid == HDFConstants.FAIL ||
            !hdf.SDfileinfo(sdid, fileInfo) )
        {
            hdf.SDend(sdid);
            return;
        }

        int sdsDatasetNumber    = fileInfo[0];
        int sdsGlobalAttrNumber = fileInfo[1];

        if (sdsDatasetNumber <= 0)
        {
            hdf.SDend(sdid);
            return;
        }

        // HDF SDS object root node
        node = new HDFObjectNode();
        node.setObjType(node.SDS);
        node.setName("SDS Array (lone)");
        node.setHasChild(true);
        parentNode = node;
        Queue sdsQueue = new Queue();
        parentNode.setChildObj(sdsQueue);

        // for SDS Global Attributes
        if (sdsGlobalAttrNumber > 0)
        {
            // HDF SDS Global Attribute object root node
            node = new HDFObjectNode();
            node.setObjType(node.SDSGLOBALATTR);
            node.setObjLevel(1);
            node.setName("Global Attributes");
            node.setHasParent(true);
            node.setParentObj(parentNode);
            node.information = "Number of attributes : "+sdsGlobalAttrNumber;
            sdsQueue.put(node);

        } // end of if (sdsGlobalAttrNumber > 0)

        // first check if the data is in the queue, if not, add it into the queue
        for (int i=0; i<sdsDatasetNumber; i++)
        {
            int sdsid = HDFConstants.FAIL;

            if ((sdsid = hdf.SDselect(sdid, i)) == HDFConstants.FAIL)
                continue;

            boolean find = find(HDFConstants.DFTAG_NDG,hdf.SDidtoref(sdsid),objQueue);
            hdf.SDendaccess(sdsid);

            if (find) continue;

            if (getSDSObject(parentNode, sdid, i, 1, sdsQueue))
                numberOfData++;
        }

        if (numberOfData > 0)
        {
            objQueue.put(parentNode);
            parentNode.information = "\nNumber of lone data sets : "+numberOfData;
        }

        hdf.SDend(sdid);

    }

    /** put the GR object into HDF queue
     * @param fid  HDF file identifier by Hopen
     * @param objQueue the queue to store the HDF object node
     *
     *  @exception ncsa.hdf.hdflib.HDFException
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     * @see HDFObjectNode
     */
    public void getGRObject(int fid , Queue objQueue)
        throws HDFException
    {
        int tag, ref;
        String info = "";
        int numberOfData = 0;
        HDFObjectNode   parentNode, childNode, node;
        int grid = hdf.GRstart(fid);
        int[] argv = new int[2];

        if ( grid==HDFConstants.FAIL || !hdf.GRfileinfo(grid,argv) )
        {
            hdf.GRend(grid);
            return;
        }

        int grDatasetNumber = argv[0];
        int grGlobalAttrNumber = argv[1];

        if (grDatasetNumber <=0 )
        {
            hdf.GRend(grid);
            return;
        }

        // HDF GR object root node
        node = new HDFObjectNode();
        node.setObjType(node.GR);
        node.setName("Raster Image (lone)");
        node.setHasChild(true);
        parentNode = node;
        Queue grQueue = new Queue();
        parentNode.setChildObj(grQueue);

        // for GR Global Attributes
        if (grGlobalAttrNumber > 0)
        {
            // add GR Global Attribute subnode
            node = new HDFObjectNode();
            node.setObjType(node.GRGLOBALATTR);
            node.setObjLevel(1);
            node.setName("Global Attributes");
            node.setHasParent(true);
            node.setParentObj(parentNode);
            node.information = "Number of attributes : "+grGlobalAttrNumber;
            grQueue.put(node);
        }

        // for GR Datasets
        for (int i=0; i<grDatasetNumber; i++)
        {
            int riid = HDFConstants.FAIL;

            if ((riid = hdf.GRselect(grid, i)) == HDFConstants.FAIL)
                continue;

            boolean find = find(HDFConstants.DFTAG_RIG, hdf.GRidtoref(riid), objQueue);
            hdf.GRendaccess(riid);

            if (find) continue;

            if (getGRObject(parentNode, grid, i, 1, grQueue))
                numberOfData++;
        }

        if (numberOfData > 0)
        {
            objQueue.put(parentNode);
            parentNode.information = "\nNumber of lone raster images : "+numberOfData;
        }

        hdf.GRend(grid);
    }

    /** put the SDS object into HDF queue
     * @param parentNode     the parent HDF object node
     * @param SDS id number  the HDF file name
     * @param Object index   the reference number of the SDS;
     * @param level          the hierachal level
     * @param sdsQueue       the queue to store the SDS node
     *
     *  @exception ncsa.hdf.hdflib.HDFException
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     * @see HDFObjectNode
     */
    public boolean getSDSObject(HDFObjectNode parentNode, int sdid,
        int index, int level, Queue sdsQueue)
        throws HDFException
    {
        //System.out.println("HDFAnalyse.getSDSObject()");

        int tag, ref;
        String info = "";
        Queue sdsDatasetQueue = new Queue();
        HDFObjectNode datasetParentNode, childNode, node;
        int[] fileInfo = new int[2];
        int sdsid = HDFConstants.FAIL;

        if ( (sdsid=hdf.SDselect(sdid,index)) == HDFConstants.FAIL)
            return false;

        tag = HDFConstants.DFTAG_NDG;
        ref = hdf.SDidtoref(sdsid);

        // check if the given SDS has dimension metadata
        // Coordinate variables are not displayed. They are created to store
        // metadata associated with dimensions. To ensure compatibility with
        // netCDF, coordinate variables are implemented as data sets
        if (hdf.SDiscoordvar(sdsid))
            return false;

        // get datsset info.
        int [] SDInfo = new int[3];
        String datasetname = new String(" ");
        String[] ss = new String[1];
        ss[0] = datasetname;
        int  dimsize[] = new int[10];
        if (!hdf.SDgetinfo(sdsid, ss, dimsize, SDInfo))
            return false;
        datasetname = ss[0];

        // get full name of the data set
        int long_name_idx = hdf.SDfindattr(sdsid, "long_name");
        int [] SDattrInfo = new int[2];
        String name = new String(" ");
        ss[0] = name;
        hdf.SDattrinfo(sdsid, long_name_idx, ss,SDattrInfo);
        name = ss[0];
        if (long_name_idx != HDFConstants.FAIL)
        {
            String long_name = new String(" ");
            int attrSize = hdf.DFKNTsize(SDattrInfo[0])*SDattrInfo[1];
            byte[] buf = new byte[attrSize];
            if (hdf.SDreadattr(sdsid, long_name_idx, buf))
            {
                long_name = new String(buf);
                datasetname += ": "+long_name;
            }
        }

        // add node information
        String rankInfo = SDInfo[0] + ", ";
        for (int j=(SDInfo[0] - 1); j>0; j--)
        {
            rankInfo += dimsize[j];
            rankInfo += " X ";
        }
        rankInfo += dimsize[0];
        info  = "\nData Name            : "+datasetname;
        info += "\nTag number           : "+tag;
        info += "\nReference number     : "+ref;
        info += "\nData type            : "+hdf.HDgetNTdesc(SDInfo[1]);
        info += "\nRank and dimensions  : "+rankInfo;
        info += "\nNumber of attributes : " + SDInfo[2];

        // add SDS dataset subnode
        node = new HDFObjectNode();
        node.setObjType(node.SDSDATASET);
        node.setObjLevel(level);
        node.setName(datasetname);
        node.setParentObj(parentNode);
        datasetParentNode = node;
        node.setTagRef(tag,ref);
        node.information = info;
        sdsQueue.put(node);

        boolean hasInfoNode = ( (ANgetobjectlabel(tag,ref)).length() > 0 ||
            (ANgetobjectdesc(tag,ref)).length() > 0 || SDInfo[2] > 0 );

        // SDS has more one components
        if (hasInfoNode)
        {
            sdsDatasetQueue = new Queue();
            node.setHasParent(true);
            node.setChildObj(sdsDatasetQueue);
            node = new HDFObjectNode();
            node.setTagRef(tag,ref);
            node.setObjType(node.SDSDATASET);
            node.setObjLevel(level+1);
            node.setIndex(index);
            node.setName(datasetname);
            node.setHasParent(true);
            node.setParentObj(datasetParentNode);
            node.information = info;
            sdsDatasetQueue.put(node);
        }

        // Date: 6-12-97:
        // Added following code to handle object annotation. but it is limited.
        // For SDS, take "Number Dataset Group(720)" as an example
        // for dataset object annotation
        if ( ( (ANgetobjectlabel(tag,ref)).length() > 0 ) ||
             ( (ANgetobjectdesc(tag,ref)).length() > 0 )   )
        {
            // HDF SDS dataset annotation   node
            node = new HDFObjectNode();
            node.setObjType(node.SDSDATASETAN);
            node.setObjLevel(level+1);
            node.setTagRef(tag,ref);
            node.setName("Dataset Annotation");
            node.setHasParent(true);
            node.setParentObj(datasetParentNode);

            // Add gr dataset node
            sdsDatasetQueue.put(node);
        }

        // for dataset attributes
        int datasetAttrNumber = SDInfo[2];
        if (datasetAttrNumber > 0)
        {
            // HDF SDS dataset Attribute object   node
            node = new HDFObjectNode();
            node.setObjType(node.SDSDATASETATTR);
            node.setObjLevel(level+1);
            node.setName("Dataset Attributes");
            node.setHasParent(true);
            node.setParentObj(datasetParentNode);
            node.information = "Number of attributes : "+datasetAttrNumber;
            node.setIndex(index);
            sdsDatasetQueue.put(node);
        } // if (datasetAttrNumber > 0)

        hdf.SDendaccess(sdsid);

        return true;
    }

    /**
     * put the GR object into HDF queue
     * @param parentNode  the parent HDF object node
     * @param grid        the GR id number
     * @param index       the index of the GR object;
     * @param level       the hierachal level
     * @param grQueue     the queue to store the HDF object node
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     * @see HDFObjectNode
     */
    public boolean getGRObject(HDFObjectNode parentNode, int grid,
        int index, int level, Queue grQueue)
        throws HDFException
    {
        int tag, ref;
        String info = "";
        Queue grDatasetQueue = new Queue();
        HDFObjectNode datasetParentNode, childNode, node;
        int riid = HDFConstants.FAIL;

        // select dataset
        if ((riid = hdf.GRselect(grid, index)) == HDFConstants.FAIL)
            return false;

        // get GR dataset information
        tag = HDFConstants.DFTAG_RIG;
        ref = hdf.GRidtoref(riid);
        int  datasetinfo[] = new int[4];
        String datasetname[] = {""};
        int  dimsize[] = new int[10];

        if (!hdf.GRgetiminfo(riid,datasetname,datasetinfo,dimsize))
        {
            hdf.GRendaccess(riid);
            return false;
        }

        String rankInfo = dimsize[0] + " X " + dimsize[1];
        info  = "\nImage Name           : "+datasetname[0];
        info += "\nTag number           : "+tag;
        info += "\nReference number     : "+ref;
        info += "\nData type            : "+hdf.HDgetNTdesc(datasetinfo[1]);
        info += "\nImage size           : "+rankInfo;
        info += "\nNumber of attributes : "+datasetinfo[3];

        // add GR dataset subnode
        node = new HDFObjectNode();
        node.setObjType(node.GRDATASET);
        node.setObjLevel(level);
        node.setName(datasetname[0]);
        node.setHasParent(true);
        node.setParentObj(parentNode);
        datasetParentNode = node;
        node.setTagRef(tag,ref);
        node.information = info;
        grQueue.put(node);

        boolean hasInfoData = ( (ANgetobjectlabel(tag,ref)).length() > 0 ||
            (ANgetobjectdesc(tag,ref)).length() > 0 || datasetinfo[3] >0 );

        // The GR has more than one compoents
        if (hasInfoData)
        {
            grDatasetQueue = new Queue();
            node.setHasChild(true);
            node.setChildObj(grDatasetQueue);
            node = new HDFObjectNode();
            node.setTagRef(tag,ref);
            node.setObjType(node.GRDATASET);
            node.setObjLevel(level+1);
            node.setIndex(index);
            node.setName(datasetname[0]);
            node.setHasParent(true);
            node.setParentObj(datasetParentNode);
            node.information = info;
            grDatasetQueue.put(node);
        }
        // Date: 6-12-97:
        // Added following code to handle object annotation. but it is limited.
        // For R8, R24 take 306 as tag for object annotation.
        // For GR, take "Raster Image Group(306)" as an example
        // for dataset object annotation
        if ( ( (ANgetobjectlabel(tag,ref)).length() > 0 ) ||
           ( (ANgetobjectdesc(tag,ref)).length() > 0 ) )
        {
            // HDF GR dataset annotation   node
            node = new HDFObjectNode();
            node.setObjType(node.GRDATASETAN);
            node.setObjLevel(level+1);
            node.setTagRef(tag,ref);
            node.setName("GR Dataset Annotation");
            node.setHasParent(true);
            node.setParentObj(datasetParentNode);
            grDatasetQueue.put(node);
        }

        // for dataset attributes
        int datasetAttrNumber = datasetinfo[3];
        if (datasetAttrNumber > 0)
        {
            // HDF GR dataset Attribute object   node
            node = new HDFObjectNode();
            node.setObjType(node.GRDATASETATTR);
            node.setObjLevel(level+1);
            node.setIndex(index);
            node.setName("Dataset Attributes");
            node.setHasParent(true);
            node.setParentObj(datasetParentNode);
            node.information = "Number of attributes : "+datasetAttrNumber;
            grDatasetQueue.put(node);
        }
        hdf.GRendaccess(riid);

        return true;
    }

    /**  insert the HDF object belongs to specified HDF Vgroup
     * @param pnode  the HDF object node.
     * @param subQueue the queue to store the HDF object belongs to Vgroup
     * @param fid    the HDF file id.
     * @param tags   the tags value
     * @param refs    the reference number
     * @param npairs  the number of objects belongs to specified Vgroup
     * @param level   the level in the HDF Object
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     */
    public  void insertVg(HDFObjectNode pnode, Queue subqueue,
        int fid, int[] tags, int[] refs, int npairs, int level)
        throws HDFException
    {
        HDFObjectNode node;
        int tag, ref;
        String info = "";

        if (npairs <= 0) return;

        for (int i=0; i<npairs; i++)
        {
            tag = tags[i];
            ref = refs[i];

            switch(tag)
            {
                case HDFConstants.DFTAG_RIG:
                case HDFConstants.DFTAG_RI:
                case HDFConstants.DFTAG_RI8:
                    // Consider only Raster Image Group, all raster
                    // images (8-bit, 24-bit, etc) have this same tag
                    int grid = hdf.GRstart(fid);
                    if ( grid != HDFConstants.FAIL)
                    {
                        int index = hdf.GRreftoindex(grid, (short)ref);
                        getGRObject(pnode, grid, index, level, subqueue);
                    }
                    hdf.GRend(grid);
                    break;
                case HDFConstants.DFTAG_SD:
                case HDFConstants.DFTAG_SDG:
                case HDFConstants.DFTAG_NDG:
                    // Scientific data, Scientific data group, Numeric data group
                    int sdid = hdf.SDstart(fileName, HDFConstants.DFACC_RDONLY);
                    if ( sdid != HDFConstants.FAIL)
                    {
                        int index = hdf.SDreftoindex(sdid, ref);
                        getSDSObject(pnode, sdid, index , level, subqueue);
                    }
                    hdf.SDend(sdid);
                    break;
                case HDFConstants.DFTAG_VH:
                case HDFConstants.DFTAG_VS:
                    // vdata description and Vdata
                    int vdataid = hdf.VSattach(fid, ref, "r");
                    String[] vdName = {""};
                    hdf.VSgetname(vdataid,vdName);

                    info  = "\nVdata Name        : "+vdName[0];
                    info += "\nTag number        : "+tag;
                    info += "\nReference number  : "+ref;
                    info += "\nNumber of records : "+hdf.VSelts(vdataid);
                    info += "\nNumber of fields  : "+hdf.VFnfields(vdataid);
                    
                    node = new HDFObjectNode();
                    node.setObjLevel(level);
                    node.setTagRef(tag,ref);
                    node.setName(vdName[0]);
                    node.setObjType(HDFObjectNode.Vdata);
                    node.setHasParent(true);
                    node.setParentObj(pnode);
                    node.information = info;
                    subqueue.put(node);
                    hdf.VSdetach(vdataid);
                    break;
                case HDFConstants.DFTAG_VG:
                    // vgroup, recursively sub-vgroup objects
                    int vgroup_id = hdf.Vattach(fid, ref, "r");
                    if (vgroup_id != HDFConstants.FAIL)
                    {
                        String vgname[] = {""};
                        hdf.Vgetname(vgroup_id, vgname);
                        int subtags[] = new int[10];
                        int subrefs[] = new int[10];
                        int subnpairs = hdf.Vgettagrefs(vgroup_id, subtags, subrefs, 10);

                        info  = "\nVgroup Name            : "+vgname[0];
                        info += "\nTag number             : "+tag;
                        info += "\nReference number       : "+ref;
                        info += "\nNumber of data objects : "+subnpairs;

                        node = new HDFObjectNode();
                        node.setName(vgname[0]);
                        node.setObjLevel(level);
                        node.setTagRef(tag,ref);
                        node.setObjType(HDFObjectNode.Vgroup);
                        node.setHasParent(true);
                        node.setParentObj(pnode);
                        subqueue.put(node);

                        if (subnpairs > 0)
                        {
                            Queue subsubqueue = new Queue(subnpairs);
                            node.setChildObj(subsubqueue);
                            node.setHasChild(true);
                            insertVg(node, subsubqueue, fid, subtags, subrefs, subnpairs, level+1);
                        }
                        hdf.Vdetach(vgroup_id);
                    }
                    break;
                default:
            } // end of switch(tag)
        } // end of for (int i=0; i<npairs; i++)
    }

    /** Does the HDF file have file label or file description.
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     * @return true if the HDF file contains file label or desc. , 
     *         false otherwise
     */
    public boolean  ANhasFileAnn(int fid)
        throws HDFException
    {
        int anid = hdf.ANstart(fid);
        boolean retValue = false;

        int fileInfo[] = new int[4];
        fileInfo[0] = fileInfo[1] = 0;

        if ( (anid != HDFConstants.FAIL)     &&
              hdf.ANfileinfo(anid, fileInfo) &&
             ((fileInfo[0] + fileInfo[1])>0))
            retValue = true;

        hdf.ANend(anid);
        return retValue;
    }

    /** get object  label from HDF file
     * @param tag  the tag of the HDF object.
     * @param ref  the reference number
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     * @return the string look like text of the object label 
     */
    public String ANgetobjectlabel(int tag,int ref)
        throws HDFException
    {
        StringBuffer retStr = new StringBuffer();
        int[] fileInfo = new int[4];
        int an_id =  hdf.ANstart(this.fid);

        if (hdf.ANfileinfo(an_id, fileInfo) == false )
        {
            hdf.ANend(an_id);
            return  retStr.toString();
        }

        int anntype = HDFConstants.AN_DATA_LABEL;
        int numObjectLabel = hdf.ANnumann(an_id, anntype, (short)tag, (short)ref);
        if (numObjectLabel == HDFConstants.FAIL) return "";
        int ann_id_list[] = new int[numObjectLabel];
        int numAnnid = hdf.ANannlist(an_id, anntype, tag, ref, ann_id_list);
        if (numAnnid == HDFConstants.FAIL) return "";
	
        for (int i=0; i<numObjectLabel; i++)
        {
            int annid = ann_id_list[i];
            int objectLabelLen = hdf.ANannlen(annid);
            if (objectLabelLen >0)
            {
                String objectLabelBuf[] = new String[1];
                objectLabelBuf[0] = new String("");
                if (hdf.ANreadann(annid, objectLabelBuf, objectLabelLen))
                {
                    retStr.append("  ");
                    retStr.append((new String(objectLabelBuf[0])).trim());
                    retStr.append("           \n");
                }
            } // if (objectLabelLen > 0)
        }  // for ()

        hdf.ANend(an_id);
        return retStr.toString();

    }

    /** get the object desc from HDF file
     * @param tag  the tag of the HDF object.
     * @param ref  the reference number
     * @return the string look like text of the object desc 
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     */
    public String ANgetobjectdesc(int tag,int ref)
        throws HDFException
    {
        StringBuffer retStr = new StringBuffer();
        int[] fileInfo = new int[4];
        int anntype = HDFConstants.AN_DATA_DESC;
        int an_id =  hdf.ANstart(this.fid);

        if (hdf.ANfileinfo(an_id, fileInfo) == false )
        {
            hdf.ANend(an_id);
            return  "";
        }

        int numObjectDesc = hdf.ANnumann(an_id, anntype, (short)tag, (short)ref);
        if (numObjectDesc == HDFConstants.FAIL) return "";
	int ann_id_list[] = new int[numObjectDesc];
	int numAnnid = hdf.ANannlist(an_id, anntype, tag, ref, ann_id_list);
	if (numAnnid == HDFConstants.FAIL) return"";

        for (int i=0; i<numObjectDesc; i++)
        {
            int annid = ann_id_list[i];
            int objectDescLen = hdf.ANannlen(annid);
            if (objectDescLen >0)
            {
                String objectDescBuf[] = new String[1];
                objectDescBuf[0] = new String("");
                if (hdf.ANreadann(annid, objectDescBuf, objectDescLen))
                {
                    retStr.append("  ");
                    retStr.append((new String(objectDescBuf[0])).trim());
                    retStr.append("   \n");
                }
            } // if (objectDescLen > 0)
        }  // for ()
	
        hdf.ANend(an_id);
        return retStr.toString();
    }

    /**
     *  Test if an data object exists in the queue by
     *  given the tag and ref of an object
     *
     *  @param tag       the tag number of the data object
     *  @param ref       the ref number of the data object
     *  @param queue     the object queue
     *  @return          true if find the object, otherwise false
     */
    public boolean find (int tag, int ref, Queue queue)
    {
        boolean find = false;
        HDFObjectNode node = null;
        Queue subQueue = null;

        if (queue.size() <= 0)
            return false;

        for (int i=0; i<queue.size(); i++)
        {
            node = (HDFObjectNode)queue.elementAt(i);
            if (node.tag == tag && node.ref == ref)
                find = true;

            if (node.child!=null && !find)
            {
                subQueue = (Queue) (node.child);
                find = find(tag, ref, subQueue);
            }
        }

        return find;
    }

    /** set the display mode to debug */
    public void setDebug() {debug = true; }

}
