/****************************************************************************
 * NCSA HDF                                                                 *
 * National Comptational Science Alliance                                   *
 * University of Illinois at Urbana-Champaign                               *
 * 605 E. Springfield, Champaign IL 61820                                   *
 *                                                                          *
 * For conditions of distribution and use, see the accompanying             *
 * hdf/COPYING file.                                                        *
 *                                                                          *
 ****************************************************************************/

package ncsa.hdf.message;

import ncsa.hdf.hdflib.HDFException;
import ncsa.hdf.hdflib.HDFLibrary;
import ncsa.hdf.hdflib.HDFConstants;

/**
 *  HDFSDS holds HDF Vdata information
 
 *  @version   1.1.3 September 22 1997
 *  @author    Peter X. Cao
 */
public class HDFVdata extends HDFObject
{
    /**
     *  the range of records
     *  <pre>
     *   records[0] = the starting index of records
     *   records[1] = the ending index of records
     *  </pre>
     */
    private int[] records;

    /** the fields array */
    private int[] fields;

    /** the arry of field names */
    private String[] fieldNames;

    /** the array of field orders */
    private int[] fieldOrders;

    /** the array of field types */
    private int[] fieldTypes;

    /** the 2-D array of vdata */
    //private String[][] vdata;

    /** the vdata name */
    private String vdataName;

    /**
     *  Constructs an HDFVdata with default parameters
     */
    public HDFVdata () { this(null, ""); }

    /**
     *  Construct an HDFVdata with specified HDFObjectNode and filename
     *
     *  @param node     the HDFObjectNode to be transfered to the server
     *  @param filename the requested HDF file
     */
    public HDFVdata (HDFObjectNode node, String filename)
    {
        super(node, filename);
        records = null;
        fields = null;
        fieldNames = null;
        fieldOrders = null;
        fieldTypes = null;
        vdataName = null;
    }

    /**
     *  create a message for sending to the client
     *
     *  @return  the HDFMessage created by this object
     */
    public HDFMessage toClient()
    {
        HDFMessage message = new HDFMessage(getClass().getName());

        message.put("information", information);
        if ( records != null ) message.put("records", records);
        if ( fields != null ) message.put("fields", fields);
        if ( fieldNames != null ) message.put("fieldNames", fieldNames);
        if ( fieldOrders != null ) message.put("fieldOrders", fieldOrders);
        if ( fieldTypes != null ) message.put("fieldTypes", fieldTypes);
        if ( data != null ) message.put("data", data);
        if ( vdataName != null ) message.put("vdataName", vdataName);
         return message;
    }

    /**
     *  serve the client request on the server
     */
    public void service()
    {
        this.hdf =  new HDFLibrary();
        try { information = getVdata (filename, nodeObject); }
        catch (Exception e) { information = "ERROR: exception in HDFVdata.getVdata()"; }
    }

    /**
     *  process the message received from the server
     *
     *  @param message the message from the server
     */
    public void fromServer(HDFMessage message)
    {
        information = (String) message.get("information");
        records = (int[]) message.get("records");
        fields = (int[]) message.get("fields");
        fieldNames = (String[]) message.get("fieldNames");
        fieldOrders = (int[]) message.get("fieldOrders");
        fieldTypes = (int[]) message.get("fieldTypes");
        data = (Object) message.get("data");
        vdataName = (String) message.get("vdataName");
    }

    /**
     *  get the selected HDF Vdata (vertex data) image for the selected HDF file
     *
     *  @param filename   the string of the hdf file name
     *  @param node       the HDFObjectNode
     *  @return           the error message
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     */
    private String getVdata (String filename, HDFObjectNode node)
        throws HDFException
    {
        int        request     = node.type;
        int        ref         = node.ref;
        int        attrIndex   = node.index;
        int fid = -1;
        String info = "";
        String vdata[][] = null;

        if ((fid = hdf.Hopen(filename)) <= 0)
        {
            return ("ERROR: Open " + filename + " failed.");
        }

        hdf.Vstart(fid);
      
        // Attaches to an existing Vdata
        int vdata_id = hdf.VSattach(fid, ref, "r");
        if (vdata_id == HDFConstants.FAIL)
        {
            hdf.Vend(fid);
            hdf.Hclose(fid);
            return ("ERROR: native call to HDFLibrary.VSattach() failed");
        }
    
        // set selected vdata field & record number
        int[] selectedRecords = new int[2];
        selectedRecords[0] = 1;
        selectedRecords[1] = hdf.VSelts(vdata_id);
        if ( selectedRecords[1] == HDFConstants.FAIL )
        {
            hdf.VSdetach(vdata_id);
            hdf.Vend(fid);
            hdf.Hclose(fid);
            return ("ERROR: native call to HDFLibrary.VSelts() failed");
        }
            
        // field
        int nfields = hdf.VFnfields(vdata_id);
        if (nfields == HDFConstants.FAIL)
        {
            hdf.VSdetach(vdata_id);
            hdf.Vend(fid);
            hdf.Hclose(fid);
            return ("ERROR: native call to HDFLibrary.VFnfields() failed");
        }

        int[] selectedFields = new int[nfields];
        for (int i=0; i < nfields; i++) 
            selectedFields[i] = i;

        info = getSpreadsheetData(vdata_id, selectedRecords, selectedFields);

        hdf.VSdetach(vdata_id);
        hdf.Vend(fid);
        hdf.Hclose(fid);
        return info;
    }

   /**
     *  get the spreadsheet data for the selected Vdata
     *
     *  @param  vdata_id   The id number of the Vdata
     *  @param  records    The selected Vdata records
     *  @param  fields     The selected Vdata fields
     *  @return            The error message
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     */
    private String getSpreadsheetData(int vdata_id, int[] records, int[] fields)
        throws HDFException
    {
        int numberOfFields = fields.length;
        int numberOfRecords = records[1]-records[0]+1;
        String[][] vdata = new String[numberOfRecords][numberOfFields];
        String[] vdataFieldName = new String[numberOfFields];
        int[] vdataFieldOrder = new int[numberOfFields];
        int[] vdataFieldType = new int[numberOfFields];
        String vdataName= "";
        Object[] data = new Object[numberOfFields];

        // get the vdata name
        String[] s = new String[1];
        s[0] = vdataName;
        hdf.VSgetname(vdata_id,s);
        vdataName = s[0];

        // get fields names, types and orders
        for (int i=0; i < numberOfFields; i++)
        {
            vdataFieldName[i] = new String(hdf.VFfieldname(vdata_id,i));
            vdataFieldType[i] = hdf.VFfieldtype(vdata_id,i);
            vdataFieldOrder[i] = hdf.VFfieldorder(vdata_id,i);
        }

        String dataStr="";
        boolean status = false;
        int  rows=0, dataType, order, datasize, dispNumber;

	// for each field of a vdata
	for (int i=0; i< numberOfFields; i++)
        {
            dataType = vdataFieldType[i];
            order = vdataFieldOrder[i];

            if ((dataType & HDFConstants.DFNT_LITEND) != 0)
                dataType -= HDFConstants.DFNT_LITEND;

            datasize = 1;
            dataStr="";

            if (order>0)
            {
                // display only the first three elements in an array;
                //if ((order>3) && (dataType != HDFConstants.DFNT_CHAR8))
                //   dispNumber = 3;
                //else dispNumber = order;

                hdf.VSseek(vdata_id, 0);

                // Specify the fields to be accessed
                status = hdf.VSsetfields(vdata_id, vdataFieldName[i]);

                //datasize = order*hdf.DFKNTsize(dataType)*numberOfRecords;
                datasize = order*numberOfRecords;

                // organize data
                int interlace = HDFConstants.FULL_INTERLACE;
                switch(dataType)
                {
                    // string
                    case HDFConstants.DFNT_CHAR:
                    case HDFConstants.DFNT_UCHAR8:
                        /*
                        byte[] bdat = new byte[datasize];
                        readNum = hdf.VSread(vdata_id, bdat, numberOfRecords, interlace);
                        rows=0;
                        for (int kk=records[0]; kk<=records[1]; kk++, rows++)
                            vdata[rows][i] = new String(bdat,order*rows,order);
                        */
                    case HDFConstants.DFNT_UINT8:
                    case HDFConstants.DFNT_INT8:
                        byte[] bdata = new byte[datasize];
                        hdf.VSread(vdata_id, (Object)bdata, numberOfRecords, interlace);
                        data[i] = bdata;
                        break;

                    /*
                    case HDFConstants.DFNT_UINT8:
                    case HDFConstants.DFNT_INT8:
                        byte[] sbdat = new byte[datasize];
                        readNum = hdf.VSread(vdata_id, sbdat, numberOfRecords,interlace);
                        rows=0;
                        for (int kk=records[0]; kk<=records[1]; kk++, rows++)
                        {
                            dataStr="";
                            for (int j=0; j<dispNumber; j++)
                            {
                                int tmpVal = (int)((byte)sbdat[j+rows*order]);
                                if ((dataType == HDFConstants.DFNT_UINT8) && (tmpVal <0))
                                    tmpVal += 256;
                                dataStr += Integer.toString(tmpVal) +"  ";
                            }

                            if (vdataFieldOrder[i]>dispNumber)
                                dataStr += "...";
                            vdata[rows][i] = new String(dataStr);
                        }
                    */
                    case HDFConstants.DFNT_INT16:
                    case HDFConstants.DFNT_UINT16:
                        /*
                        short[] sdat = new short[order*numberOfRecords];
                        readNum = hdf.VSread(vdata_id, sdat, numberOfRecords, interlace);
                        rows=0;
                        for (int kk=records[0]; kk<=records[1]; kk++, rows++)
                        {
                            dataStr="";
                            int pos = 0;
                            for (int j=0; j<dispNumber; j++)
                            {
                                int tmpVal = (int)sdat[j+order*rows];
                                if ((dataType == HDFConstants.DFNT_UINT16) && (tmpVal <0))
                                    tmpVal += 65536;
                                dataStr += Integer.toString(tmpVal) +"  ";
                            }

                            if (vdataFieldOrder[i]>dispNumber)
                                dataStr += "...";

                            vdata[rows][i] = new String(dataStr);
                        }
                        */
                        short[] sdata = new short[datasize];
                        hdf.VSread(vdata_id, (Object)sdata, numberOfRecords,interlace);
                        data[i] = sdata;
                        break;

                    case HDFConstants.DFNT_INT32:
                    case HDFConstants.DFNT_UINT32:
                        /*
                        dataStr="";
                        int[] idat = new int[order*numberOfRecords];
                        hdf.VSread(vdata_id, idat, numberOfRecords, interlace);
                        rows=0;
                        for (int kk=records[0]; kk<=records[1]; kk++, rows++)
                        {
                            dataStr="";
                            for (int j=0; j<dispNumber; j++)
                            {
                                int tmpVal = idat[j+order*rows];
                                if ((dataType == HDFConstants.DFNT_UINT32) && (tmpVal <0))
                                    tmpVal += 4294967296L;
                                dataStr += Integer.toString(tmpVal) +"  ";
                            }

                            if (vdataFieldOrder[i]>dispNumber)
                                dataStr += "...";

                            vdata[rows][i] = new String(dataStr);
                        }
                        */
                        int[] idata = new int[datasize];
                        hdf.VSread(vdata_id, (Object)idata, numberOfRecords,interlace);
                        data[i] = idata;
                        break;

                    case HDFConstants.DFNT_FLOAT:
                    //case HDFConstants.DFNT_FLOAT32:
                        /*
                        float[] fdat = new float[order*numberOfRecords];
                        hdf.VSread(vdata_id, fdat, numberOfRecords, interlace);
                        rows=0;
                        for (int kk=records[0]; kk<=records[1]; kk++, rows++)
                        {
                            dataStr="";
                            for (int j=0; j<dispNumber; j++)
                            {
                                float tmpVal = fdat[j+order*rows];
                                dataStr += Float.toString(tmpVal) +"  ";
                            }

                            if (vdataFieldOrder[i]>dispNumber)
                               dataStr += "...";

                            vdata[rows][i] = new String(dataStr);
                        }
                        */
                        float[] fdata = new float[datasize];
                        hdf.VSread(vdata_id, (Object)fdata, numberOfRecords,interlace);
                        data[i] = fdata;
                        break;

                    case HDFConstants.DFNT_DOUBLE:
                    //case HDFConstants.DFNT_FLOAT64:
                        /*
                        double[] ddat = new double[order*numberOfRecords];
                        hdf.VSread(vdata_id, ddat, numberOfRecords, interlace);
                        rows = 0;
                        for (int kk=records[0]; kk<=records[1]; kk++, rows++)
                        {
                            dataStr="";
                            for (int j=0; j<dispNumber; j++)
                            {
                                double tmpVal =ddat[j+order*rows];
                                dataStr += Double.toString(tmpVal) +"  ";
                            }

                            if (vdataFieldOrder[i]>dispNumber)
                               dataStr += "...";

                            vdata[rows][i] = new String(dataStr);
                        }
                        */
                        double[] ddata = new double[datasize];
                        hdf.VSread(vdata_id, (Object)ddata, numberOfRecords,interlace);
                        data[i] = ddata;
                        break;

                    default:
                        //vdata[rows][i] = new String("");
                        data[i] = new byte[1];
                }
            }
        }

        this.records = records;
        this.fields = fields;
        this.fieldNames = vdataFieldName;
        this.fieldOrders = vdataFieldOrder;
        this.fieldTypes = vdataFieldType;
        this.data = data;
        this.vdataName = vdataName;
        return "";
    }

    /** return the records range */
    public int[] getRecords() { return records; }

    /** return the Vdata fields */
    public int[] getFields() { return fields; }

    /** return the array of fields names */
    public String[] getFieldNames() { return fieldNames; }

    /** return the array of the field orders */
    public int[] getFieldOrders() { return fieldOrders; }

    /** return the array of the fields types */
    public int[] getFieldTypes() { return fieldTypes; }

    /** return the Vdata */
    //public String[][] getVdata() { return vdata; }

    /** return the name of the Vdata */
    public String getVdataName() { return vdataName; }

}
